const gulp = require('gulp');
const fse = require('fs-extra');
const path = require('path');
const ts = require('gulp-typescript');
const merge2 = require('merge2');
const through2 = require('through2');
const cwd = process.cwd();
const ES = 'es';
const LIB = 'lib';
const SRC = 'src';
const esDir = path.join(cwd, ES);
const libDir = path.join(cwd, LIB);
const tsConfigPath = path.join(cwd, 'tsconfig.json');
const tsDefaultReporter = ts.reporter.defaultReporter();
const styles = [
    'src/**/*.less',
];
const sources = [
    'src/**/*.tsx',
    'src/**/*.ts',
    'src/**/*.jsx',
];

function isStyles(value) {
    const styles = ['.less', '.stylus', '.sass'];
    return styles.includes(path.extname(value));
}

function clean(dir) {
    fse.removeSync(dir);
}

gulp.task('clean', done => {
    clean(esDir);
    clean(libDir);
    done();
});
gulp.task('compile-with-es', done => {
    compile({
        esm: true,
        single: false,
        styles,
        sources,
    }).on('finish', done);
});
gulp.task('compile-with-lib', done => {
    compile({
        esm: false,
        single: false,
        styles,
        sources,
    }).on('finish', done);
});

function replace(src, dest) {
    return through2.obj(function(file, encoding, next) {
        file.path = file.path.replace(src, dest);
        this.push(file);
        next();
    });
}

function compile({ esm, styles, sources, single }) {
    const dir = esm ? esDir : libDir;
    const module = esm ? ES : LIB;
    const dest = single ? cwd : dir;
    const streams = [];
    if (styles.length > 0) {
        const less = gulp
            .src(styles)
            .pipe(replace(SRC, module))
            .pipe(gulp.dest(dest));
        streams.push(less);
    }
    if (sources.length > 0) {
        const tsConfig = fse.readJsonSync(tsConfigPath);
        const compileOptions = Object.assign({
            noUnusedParameters: false,
            noUnusedLocals: false,
            strictNullChecks: true,
            jsx: 'preserve',
            moduleResolution: 'node',
            declaration: true,
            allowSyntheticDefaultImports: true,
        }, tsConfig.compilerOptions);
        let sourceStream = gulp.src(sources);
        if (!esm) {
            compileOptions.target = 'es5';
        }
        let hasError = false;
        const tsResult = sourceStream.pipe(ts(compileOptions, {
            error(e) {
                hasError = true;
                tsDefaultReporter.error(e);
            },
            finish: tsDefaultReporter.finish,
        }));
        const tsFilesStream = tsResult
            .js
            .pipe(replace(SRC, module))
            .pipe(gulp.dest(dest));
        const tsd = tsResult
            .dts
            .pipe(replace(SRC, module))
            .pipe(gulp.dest(dest));

        function check() {
            if (hasError) {
                process.exit(1);
            }
        }

        tsResult.on('finish', check);
        tsResult.on('end', check);
        streams.push(tsFilesStream);
        streams.push(tsd);
    }
    return merge2(streams.filter(s => s));
}

gulp.task('compile', gulp.series('clean', gulp.parallel([
    'compile-with-es',
    'compile-with-lib',
]), done => {
    done();
}));
gulp.task('watch-with-all', done => {
    function getCompileOptions(path) {
        return {
            styles: isStyles(path) ? [path] : [],
            sources: isStyles(path) ? [] : [path],
        };
    }

    const watcher = gulp.watch([...sources, ...styles]);
    watcher.on('change', function(path, stats) {
        console.log(`File ${path} was change`);
        compile(Object.assign({ esm: true, single: true }, getCompileOptions(path)));
        compile(Object.assign({ esm: false, single: true }, getCompileOptions(path)));
    });
    watcher.on('add', function(path, stats) {
        console.log(`File ${path} was added`);
    });
    watcher.on('unlink', function(path, stats) {
        console.log(`File ${path} was removed`);
    });
});
gulp.task('watch', gulp.series(gulp.parallel([
    'watch-with-all',
]), done => {
    done();
}));
